@@ -49,27 +49,41 @@
     resolved to a Python object, and otherwise it becomes simple emphasis.
     This can be used as the default role to make links 'smart'.
 
-    :copyright: Copyright 2007-2016 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2018 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
+import inspect
 import os
+import posixpath
 import re
 import sys
-import inspect
-import posixpath
 from types import ModuleType
 
-from six import text_type
-from docutils.parsers.rst import directives
-from docutils.statemachine import ViewList
 from docutils import nodes
+from docutils.parsers.rst import Directive, directives
+from docutils.statemachine import ViewList
+from six import string_types
+from six import text_type
 
 import sphinx
 from sphinx import addnodes
-from sphinx.util.compat import Directive
+from sphinx.environment.adapters.toctree import TocTree
+from sphinx.ext.autodoc import get_documenters
+from sphinx.ext.autodoc.directive import DocumenterBridge, Options
+from sphinx.ext.autodoc.importer import import_module
 from sphinx.pycode import ModuleAnalyzer, PycodeError
-from sphinx.ext.autodoc import Options
+from sphinx.util import import_object, rst, logging
+
+if False:
+    # For type annotation
+    from typing import Any, Dict, List, Tuple, Type, Union  # NOQA
+    from docutils.utils import Inliner  # NOQA
+    from sphinx.application import Sphinx  # NOQA
+    from sphinx.environment import BuildEnvironment  # NOQA
+    from sphinx.ext.autodoc import Documenter  # NOQA
+
+logger = logging.getLogger(__name__)
 
 
 # -- autosummary_toc node ------------------------------------------------------
@@ -79,6 +93,7 @@ class autosummary_toc(nodes.comment):
 
 
 def process_autosummary_toc(app, doctree):
+    # type: (Sphinx, nodes.Node) -> None
     """Insert items described in autosummary:: to the TOC tree, but do
     not generate the toctree:: list.
     """
@@ -91,23 +106,25 @@ def process_autosummary_toc(app, doctree):
             try:
                 if (isinstance(subnode, autosummary_toc) and
                         isinstance(subnode[0], addnodes.toctree)):
-                    env.note_toctree(env.docname, subnode[0])
+                    TocTree(env).note(env.docname, subnode[0])
                     continue
             except IndexError:
                 continue
             if not isinstance(subnode, nodes.section):
                 continue
             if subnode not in crawled:
-                crawl_toc(subnode, depth+1)
+                crawl_toc(subnode, depth + 1)
     crawl_toc(doctree)
 
 
 def autosummary_toc_visit_html(self, node):
+    # type: (nodes.NodeVisitor, autosummary_toc) -> None
     """Hide autosummary toctree list in HTML output."""
     raise nodes.SkipNode
 
 
 def autosummary_noop(self, node):
+    # type: (nodes.NodeVisitor, nodes.Node) -> None
     pass
 
 
@@ -118,6 +135,7 @@ class autosummary_table(nodes.comment):
 
 
 def autosummary_table_visit_html(self, node):
+    # type: (nodes.NodeVisitor, autosummary_table) -> None
     """Make the first column of the table non-breaking."""
     try:
         tbody = node[0][0][-1]
@@ -135,12 +153,13 @@ def autosummary_table_visit_html(self, node):
 
 # -- autodoc integration -------------------------------------------------------
 
-class FakeDirective:
-    env = {}
-    genopt = Options()
+class FakeDirective(DocumenterBridge):
+    def __init__(self):
+        super(FakeDirective, self).__init__({}, None, Options(), 0)  # type: ignore
 
 
-def get_documenter(obj, parent):
+def get_documenter(app, obj, parent):
+    # type: (Sphinx, Any, Any) -> Type[Documenter]
     """Get an autodoc.Documenter class suitable for documenting the given
     object.
 
@@ -148,8 +167,7 @@ def get_documenter(obj, parent):
     another Python object (e.g. a module or a class) to which *obj*
     belongs to.
     """
-    from sphinx.ext.autodoc import AutoDirective, DataDocumenter, \
-        ModuleDocumenter
+    from sphinx.ext.autodoc import DataDocumenter, ModuleDocumenter
 
     if inspect.ismodule(obj):
         # ModuleDocumenter.can_document_member always returns False
@@ -157,7 +175,7 @@ def get_documenter(obj, parent):
 
     # Construct a fake documenter for *parent*
     if parent is not None:
-        parent_doc_cls = get_documenter(parent, None)
+        parent_doc_cls = get_documenter(app, parent, None)
     else:
         parent_doc_cls = ModuleDocumenter
 
@@ -167,7 +185,7 @@ def get_documenter(obj, parent):
         parent_doc = parent_doc_cls(FakeDirective(), "")
 
     # Get the corrent documenter class for *obj*
-    classes = [cls for cls in AutoDirective._registry.values()
+    classes = [cls for cls in get_documenters(app).values()
                if cls.can_document_member(obj, '', False, parent_doc)]
     if classes:
         classes.sort(key=lambda cls: cls.priority)
@@ -196,13 +214,15 @@ class Autosummary(Directive):
     }
 
     def warn(self, msg):
+        # type: (unicode) -> None
         self.warnings.append(self.state.document.reporter.warning(
             msg, line=self.lineno))
 
     def run(self):
+        # type: () -> List[nodes.Node]
         self.env = env = self.state.document.settings.env
         self.genopt = Options()
-        self.warnings = []
+        self.warnings = []  # type: List[nodes.Node]
         self.result = ViewList()
 
         names = [x.strip().split()[0] for x in self.content
@@ -235,6 +255,7 @@ class Autosummary(Directive):
         return self.warnings + nodes
 
     def get_items(self, names):
+        # type: (List[unicode]) -> List[Tuple[unicode, unicode, unicode, unicode]]
         """Try to import the given names, and return a list of
         ``[(name, signature, summary_string, real_name), ...]``.
         """
@@ -242,7 +263,7 @@ class Autosummary(Directive):
 
         prefixes = get_import_prefixes_from_env(env)
 
-        items = []
+        items = []  # type: List[Tuple[unicode, unicode, unicode, unicode]]
 
         max_item_chars = 50
 
@@ -264,10 +285,10 @@ class Autosummary(Directive):
             if not isinstance(obj, ModuleType):
                 # give explicitly separated module name, so that members
                 # of inner classes can be documented
-                full_name = modname + '::' + full_name[len(modname)+1:]
+                full_name = modname + '::' + full_name[len(modname) + 1:]
             # NB. using full_name here is important, since Documenters
             #     handle module prefixes slightly differently
-            documenter = get_documenter(obj, parent)(self, full_name)
+            documenter = get_documenter(self.env.app, obj, parent)(self, full_name)
             if not documenter.parse_name():
                 self.warn('failed to parse name %s' % real_name)
                 items.append((display_name, '', '', real_name))
@@ -287,8 +308,7 @@ class Autosummary(Directive):
                 # be cached anyway)
                 documenter.analyzer.find_attr_docs()
             except PycodeError as err:
-                documenter.env.app.debug(
-                    '[autodoc] module analyzer failed: %s', err)
+                logger.debug('[autodoc] module analyzer failed: %s', err)
                 # no source file -- e.g. for builtin and C modules
                 documenter.analyzer = None
 
@@ -300,12 +320,11 @@ class Autosummary(Directive):
             else:
                 max_chars = max(10, max_item_chars - len(display_name))
                 sig = mangle_signature(sig, max_chars=max_chars)
-                sig = sig.replace('*', r'\*')
 
             # -- Grab the summary
 
             documenter.add_content(None)
-            doc = list(documenter.process_doc([self.result.data]))
+            doc = self.result.data
 
             while doc and not doc[0].strip():
                 doc.pop(0)
@@ -332,12 +351,13 @@ class Autosummary(Directive):
         return items
 
     def get_table(self, items):
+        # type: (List[Tuple[unicode, unicode, unicode, unicode]]) -> List[Union[addnodes.tabular_col_spec, autosummary_table]]  # NOQA
         """Generate a proper list of table nodes for autosummary:: directive.
 
         *items* is a list produced by :meth:`get_items`.
         """
         table_spec = addnodes.tabular_col_spec()
-        table_spec['spec'] = 'p{0.5\linewidth}p{0.5\linewidth}'
+        table_spec['spec'] = r'p{0.5\linewidth}p{0.5\linewidth}'
 
         table = autosummary_table('')
         real_table = nodes.table('', classes=['longtable'])
@@ -350,6 +370,7 @@ class Autosummary(Directive):
         group.append(body)
 
         def append_row(*column_texts):
+            # type: (unicode) -> None
             row = nodes.row('')
             for text in column_texts:
                 node = nodes.paragraph('')
@@ -367,7 +388,7 @@ class Autosummary(Directive):
         for name, sig, summary, real_name in items:
             qualifier = 'obj'
             if 'nosignatures' not in self.options:
-                col1 = ':%s:`%s <%s>`\ %s' % (qualifier, name, real_name, sig)
+                col1 = ':%s:`%s <%s>`\\ %s' % (qualifier, name, real_name, rst.escape(sig))  # type: unicode  # NOQA
             else:
                 col1 = ':%s:`%s <%s>`' % (qualifier, name, real_name)
             col2 = summary
@@ -376,9 +397,20 @@ class Autosummary(Directive):
         return [table_spec, table]
 
 
+def strip_arg_typehint(s):
+    # type: (unicode) -> unicode
+    """Strip a type hint from argument definition."""
+    return s.split(':')[0].strip()
+
+
 def mangle_signature(sig, max_chars=30):
+    # type: (unicode, int) -> unicode
     """Reformat a function signature to a more compact form."""
-    s = re.sub(r"^\((.*)\)$", r"\1", sig).strip()
+    # Strip return type annotation
+    s = re.sub(r"\)\s*->\s.*$", ")", sig)
+
+    # Remove parenthesis
+    s = re.sub(r"^\((.*)\)$", r"\1", s).strip()
 
     # Strip strings (which can contain things that confuse the code below)
     s = re.sub(r"\\\\", "", s)
@@ -386,12 +418,12 @@ def mangle_signature(sig, max_chars=30):
     s = re.sub(r"'[^']*'", "", s)
 
     # Parse the signature to arguments + options
-    args = []
-    opts = []
+    args = []  # type: List[unicode]
+    opts = []  # type: List[unicode]
 
     opt_re = re.compile(r"^(.*, |)([a-zA-Z0-9_*]+)=")
     while s:
-        m = opt_re.search(s)
+        m = opt_re.search(s)  # type: ignore
         if not m:
             # The rest are arguments
             args = s.split(', ')
@@ -400,19 +432,27 @@ def mangle_signature(sig, max_chars=30):
         opts.insert(0, m.group(2))
         s = m.group(1)[:-2]
 
+    # Strip typehints
+    for i, arg in enumerate(args):
+        args[i] = strip_arg_typehint(arg)
+
+    for i, opt in enumerate(opts):
+        opts[i] = strip_arg_typehint(opt)
+
     # Produce a more compact signature
-    sig = limited_join(", ", args, max_chars=max_chars-2)
+    sig = limited_join(", ", args, max_chars=max_chars - 2)
     if opts:
         if not sig:
-            sig = "[%s]" % limited_join(", ", opts, max_chars=max_chars-4)
+            sig = "[%s]" % limited_join(", ", opts, max_chars=max_chars - 4)
         elif len(sig) < max_chars - 4 - 2 - 3:
             sig += "[, %s]" % limited_join(", ", opts,
-                                           max_chars=max_chars-len(sig)-4-2)
+                                           max_chars=max_chars - len(sig) - 4 - 2)
 
     return u"(%s)" % sig
 
 
 def limited_join(sep, items, max_chars=30, overflow_marker="..."):
+    # type: (unicode, List[unicode], int, unicode) -> unicode
     """Join a number of strings to one, limiting the length to *max_chars*.
 
     If the string overflows this limit, replace the last fitting item by
@@ -439,11 +479,12 @@ def limited_join(sep, items, max_chars=30, overflow_marker="..."):
 # -- Importing items -----------------------------------------------------------
 
 def get_import_prefixes_from_env(env):
+    # type: (BuildEnvironment) -> List
     """
     Obtain current Python import prefixes (for `import_by_name`)
     from ``document.env``
     """
-    prefixes = [None]
+    prefixes = [None]  # type: List
 
     currmodule = env.ref_context.get('py:module')
     if currmodule:
@@ -460,6 +501,7 @@ def get_import_prefixes_from_env(env):
 
 
 def import_by_name(name, prefixes=[None]):
+    # type: (unicode, List) -> Tuple[unicode, Any, Any, unicode]
     """Import a Python object that has the given *name*, under one of the
     *prefixes*.  The first name that succeeds is used.
     """
@@ -478,6 +520,7 @@ def import_by_name(name, prefixes=[None]):
 
 
 def _import_by_name(name):
+    # type: (str) -> Tuple[Any, Any, unicode]
     """Import a Python object given its full name."""
     try:
         name_parts = name.split('.')
@@ -486,8 +529,7 @@ def _import_by_name(name):
         modname = '.'.join(name_parts[:-1])
         if modname:
             try:
-                __import__(modname)
-                mod = sys.modules[modname]
+                mod = import_module(modname)
                 return getattr(mod, name_parts[-1]), mod, modname
             except (ImportError, IndexError, AttributeError):
                 pass
@@ -495,13 +537,14 @@ def _import_by_name(name):
         # ... then as MODNAME, MODNAME.OBJ1, MODNAME.OBJ1.OBJ2, ...
         last_j = 0
         modname = None
-        for j in reversed(range(1, len(name_parts)+1)):
+        for j in reversed(range(1, len(name_parts) + 1)):
             last_j = j
             modname = '.'.join(name_parts[:j])
             try:
-                __import__(modname)
+                import_module(modname)
             except ImportError:
                 continue
+
             if modname in sys.modules:
                 break
 
@@ -520,14 +563,15 @@ def _import_by_name(name):
 
 # -- :autolink: (smart default role) -------------------------------------------
 
-def autolink_role(typ, rawtext, etext, lineno, inliner,
-                  options={}, content=[]):
+def autolink_role(typ, rawtext, etext, lineno, inliner, options={}, content=[]):
+    # type: (unicode, unicode, unicode, int, Inliner, Dict, List[unicode]) -> Tuple[List[nodes.Node], List[nodes.Node]]  # NOQA
     """Smart linking role.
 
     Expands to ':obj:`text`' if `text` is an object that can be imported;
     otherwise expands to '*text*'.
     """
     env = inliner.document.settings.env
+    r = None  # type: Tuple[List[nodes.Node], List[nodes.Node]]
     r = env.get_domain('py').role('obj')(
         'obj', rawtext, etext, lineno, inliner, options, content)
     pnode = r[0][0]
@@ -536,13 +580,33 @@ def autolink_role(typ, rawtext, etext, lineno, inliner,
     try:
         name, obj, parent, modname = import_by_name(pnode['reftarget'], prefixes)
     except ImportError:
-        content = pnode[0]
-        r[0][0] = nodes.emphasis(rawtext, content[0].astext(),
-                                 classes=content['classes'])
+        content_node = pnode[0]
+        r[0][0] = nodes.emphasis(rawtext, content_node[0].astext(),
+                                 classes=content_node['classes'])
     return r
 
 
+def get_rst_suffix(app):
+    # type: (Sphinx) -> unicode
+    def get_supported_format(suffix):
+        # type: (unicode) -> Tuple[unicode]
+        parser_class = app.registry.get_source_parsers().get(suffix)
+        if parser_class is None:
+            return ('restructuredtext',)
+        if isinstance(parser_class, string_types):
+            parser_class = import_object(parser_class, 'source parser')  # type: ignore
+        return parser_class.supported
+
+    suffix = None  # type: unicode
+    for suffix in app.config.source_suffix:
+        if 'restructuredtext' in get_supported_format(suffix):
+            return suffix
+
+    return None
+
+
 def process_generate_options(app):
+    # type: (Sphinx) -> None
     genfiles = app.config.autosummary_generate
 
     if genfiles and not hasattr(genfiles, '__len__'):
@@ -555,16 +619,24 @@ def process_generate_options(app):
 
     from sphinx.ext.autosummary.generate import generate_autosummary_docs
 
-    ext = app.config.source_suffix[0]
-    genfiles = [genfile + (not genfile.endswith(ext) and ext or '')
+    ext = app.config.source_suffix
+    genfiles = [genfile + (not genfile.endswith(tuple(ext)) and ext[0] or '')
                 for genfile in genfiles]
 
+    suffix = get_rst_suffix(app)
+    if suffix is None:
+        logger.warning('autosummary generats .rst files internally. '
+                       'But your source_suffix does not contain .rst. Skipped.')
+        return
+
     generate_autosummary_docs(genfiles, builder=app.builder,
-                              warn=app.warn, info=app.info, suffix=ext,
-                              base_path=app.srcdir)
+                              warn=logger.warning, info=logger.info,
+                              suffix=suffix, base_path=app.srcdir,
+                              app=app)
 
 
 def setup(app):
+    # type: (Sphinx) -> Dict[unicode, Any]
     # I need autodoc
     app.setup_extension('sphinx.ext.autodoc')
     app.add_node(autosummary_toc,
