@@ -20,112 +20,77 @@
       also be specified individually, e.g. if the docs should be buildable
       without Internet access.
 
-    :copyright: Copyright 2007-2016 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2018 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
 from __future__ import print_function
 
-import time
-import zlib
-import codecs
 import functools
 import posixpath
+import sys
+import time
 from os import path
-import re
 
-from six import iteritems, string_types
-from six.moves.urllib.parse import urlsplit, urlunsplit
 from docutils import nodes
 from docutils.utils import relative_path
+from six import PY3, iteritems, string_types
+from six.moves.urllib.parse import urlsplit, urlunsplit
 
 import sphinx
-from sphinx.locale import _
 from sphinx.builders.html import INVENTORY_FILENAME
-from sphinx.util import requests
+from sphinx.locale import _
+from sphinx.util import requests, logging
+from sphinx.util.inventory import InventoryFile
+
+if False:
+    # For type annotation
+    from typing import Any, Dict, IO, List, Tuple, Union  # NOQA
+    from sphinx.application import Sphinx  # NOQA
+    from sphinx.config import Config  # NOQA
+    from sphinx.environment import BuildEnvironment  # NOQA
+
+    if PY3:
+        unicode = str
+
+    Inventory = Dict[unicode, Dict[unicode, Tuple[unicode, unicode, unicode, unicode]]]
+
+logger = logging.getLogger(__name__)
 
 
-UTF8StreamReader = codecs.lookup('utf-8')[2]
+class InventoryAdapter(object):
+    """Inventory adapter for environment"""
 
+    def __init__(self, env):
+        self.env = env
 
-def read_inventory_v1(f, uri, join):
-    f = UTF8StreamReader(f)
-    invdata = {}
-    line = next(f)
-    projname = line.rstrip()[11:]
-    line = next(f)
-    version = line.rstrip()[11:]
-    for line in f:
-        name, type, location = line.rstrip().split(None, 2)
-        location = join(uri, location)
-        # version 1 did not add anchors to the location
-        if type == 'mod':
-            type = 'py:module'
-            location += '#module-' + name
-        else:
-            type = 'py:' + type
-            location += '#' + name
-        invdata.setdefault(type, {})[name] = (projname, version, location, '-')
-    return invdata
+        if not hasattr(env, 'intersphinx_cache'):
+            self.env.intersphinx_cache = {}
+            self.env.intersphinx_inventory = {}
+            self.env.intersphinx_named_inventory = {}
 
+    @property
+    def cache(self):
+        # type: () -> Dict[unicode, Tuple[unicode, int, Inventory]]
+        return self.env.intersphinx_cache
 
-def read_inventory_v2(f, uri, join, bufsize=16 * 1024):
-    invdata = {}
-    line = f.readline()
-    projname = line.rstrip()[11:].decode('utf-8')
-    line = f.readline()
-    version = line.rstrip()[11:].decode('utf-8')
-    line = f.readline().decode('utf-8')
-    if 'zlib' not in line:
-        raise ValueError
+    @property
+    def main_inventory(self):
+        # type: () -> Inventory
+        return self.env.intersphinx_inventory
 
-    def read_chunks():
-        decompressor = zlib.decompressobj()
-        for chunk in iter(lambda: f.read(bufsize), b''):
-            yield decompressor.decompress(chunk)
-        yield decompressor.flush()
+    @property
+    def named_inventory(self):
+        # type: () -> Dict[unicode, Inventory]
+        return self.env.intersphinx_named_inventory
 
-    def split_lines(iter):
-        buf = b''
-        for chunk in iter:
-            buf += chunk
-            lineend = buf.find(b'\n')
-            while lineend != -1:
-                yield buf[:lineend].decode('utf-8')
-                buf = buf[lineend + 1:]
-                lineend = buf.find(b'\n')
-        assert not buf
-
-    for line in split_lines(read_chunks()):
-        # be careful to handle names with embedded spaces correctly
-        m = re.match(r'(?x)(.+?)\s+(\S*:\S*)\s+(-?\d+)\s+(\S+)\s+(.*)',
-                     line.rstrip())
-        if not m:
-            continue
-        name, type, prio, location, dispname = m.groups()
-        if type == 'py:module' and type in invdata and \
-                name in invdata[type]:  # due to a bug in 1.1 and below,
-                                        # two inventory entries are created
-                                        # for Python modules, and the first
-                                        # one is correct
-            continue
-        if location.endswith(u'$'):
-            location = location[:-1] + name
-        location = join(uri, location)
-        invdata.setdefault(type, {})[name] = (projname, version,
-                                              location, dispname)
-    return invdata
-
-
-def read_inventory(f, uri, join, bufsize=16 * 1024):
-    line = f.readline().rstrip().decode('utf-8')
-    if line == '# Sphinx inventory version 1':
-        return read_inventory_v1(f, uri, join)
-    elif line == '# Sphinx inventory version 2':
-        return read_inventory_v2(f, uri, join, bufsize=bufsize)
+    def clear(self):
+        self.env.intersphinx_inventory.clear()
+        self.env.intersphinx_named_inventory.clear()
 
 
 def _strip_basic_auth(url):
+    # type: (unicode) -> unicode
     """Returns *url* with basic auth credentials removed. Also returns the
     basic auth username and password if they're present in *url*.
 
@@ -147,6 +112,7 @@ def _strip_basic_auth(url):
 
 
 def _read_from_url(url, config=None):
+    # type: (unicode, Config) -> IO
     """Reads data from *url* with an HTTP *GET*.
 
     This function supports fetching from resources which use basic HTTP auth as
@@ -172,6 +138,7 @@ def _read_from_url(url, config=None):
 
 
 def _get_safe_url(url):
+    # type: (unicode) -> unicode
     """Gets version of *url* with basic auth passwords obscured. This function
     returns results suitable for printing and logging.
 
@@ -197,6 +164,7 @@ def _get_safe_url(url):
 
 
 def fetch_inventory(app, uri, inv):
+    # type: (Sphinx, unicode, Any) -> Any
     """Fetch, parse and return an intersphinx inventory file."""
     # both *uri* (base URI of the links to generate) and *inv* (actual
     # location of the inventory file) can be local or remote URIs
@@ -210,47 +178,48 @@ def fetch_inventory(app, uri, inv):
         else:
             f = open(path.join(app.srcdir, inv), 'rb')
     except Exception as err:
-        app.warn('intersphinx inventory %r not fetchable due to '
-                 '%s: %s' % (inv, err.__class__, err))
-        return
+        err.args = ('intersphinx inventory %r not fetchable due to %s: %s',
+                    inv, err.__class__, err)
+        raise
     try:
         if hasattr(f, 'url'):
-            newinv = f.url
+            newinv = f.url  # type: ignore
             if inv != newinv:
-                app.info('intersphinx inventory has moved: %s -> %s' % (inv, newinv))
+                logger.info('intersphinx inventory has moved: %s -> %s', inv, newinv)
 
                 if uri in (inv, path.dirname(inv), path.dirname(inv) + '/'):
                     uri = path.dirname(newinv)
         with f:
             try:
                 join = localuri and path.join or posixpath.join
-                invdata = read_inventory(f, uri, join)
+                invdata = InventoryFile.load(f, uri, join)
             except ValueError as exc:
                 raise ValueError('unknown or unsupported inventory version: %r' % exc)
     except Exception as err:
-        app.warn('intersphinx inventory %r not readable due to '
-                 '%s: %s' % (inv, err.__class__.__name__, err))
+        err.args = ('intersphinx inventory %r not readable due to %s: %s',
+                    inv, err.__class__.__name__, err)
+        raise
     else:
         return invdata
 
 
 def load_mappings(app):
+    # type: (Sphinx) -> None
     """Load all intersphinx mappings into the environment."""
     now = int(time.time())
     cache_time = now - app.config.intersphinx_cache_limit * 86400
-    env = app.builder.env
-    if not hasattr(env, 'intersphinx_cache'):
-        env.intersphinx_cache = {}
-        env.intersphinx_inventory = {}
-        env.intersphinx_named_inventory = {}
-    cache = env.intersphinx_cache
+    inventories = InventoryAdapter(app.builder.env)
     update = False
     for key, value in iteritems(app.config.intersphinx_mapping):
+        name = None  # type: unicode
+        uri = None   # type: unicode
+        inv = None   # type: Union[unicode, Tuple[unicode, ...]]
+
         if isinstance(value, (list, tuple)):
             # new format
             name, (uri, inv) = key, value
             if not isinstance(name, string_types):
-                app.warn('intersphinx identifier %r is not string. Ignored' % name)
+                logger.warning('intersphinx identifier %r is not string. Ignored', name)
                 continue
         else:
             # old format, no name
@@ -261,27 +230,45 @@ def load_mappings(app):
         if not isinstance(inv, tuple):
             invs = (inv, )
         else:
-            invs = inv
+            invs = inv  # type: ignore
 
+        failures = []
         for inv in invs:
             if not inv:
                 inv = posixpath.join(uri, INVENTORY_FILENAME)
             # decide whether the inventory must be read: always read local
             # files; remote ones only if the cache time is expired
-            if '://' not in inv or uri not in cache \
-                    or cache[uri][1] < cache_time:
-                safe_inv_url = _get_safe_url(inv)
-                app.info(
-                    'loading intersphinx inventory from %s...' % safe_inv_url)
-                invdata = fetch_inventory(app, uri, inv)
+            if '://' not in inv or uri not in inventories.cache \
+                    or inventories.cache[uri][1] < cache_time:
+                safe_inv_url = _get_safe_url(inv)  # type: ignore
+                logger.info('loading intersphinx inventory from %s...', safe_inv_url)
+                try:
+                    invdata = fetch_inventory(app, uri, inv)
+                except Exception as err:
+                    failures.append(err.args)
+                    continue
+
                 if invdata:
-                    cache[uri] = (name, now, invdata)
+                    inventories.cache[uri] = (name, now, invdata)
                     update = True
                     break
 
+        if failures == []:
+            pass
+        elif len(failures) < len(invs):
+            logger.info("encountered some issues with some of the inventories,"
+                        " but they had working alternatives:")
+            for fail in failures:
+                logger.info(*fail)
+        else:
+            logger.warning("failed to reach any of the inventories "
+                           "with the following issues:")
+            for fail in failures:
+                logger.warning(*fail)
+
     if update:
-        env.intersphinx_inventory = {}
-        env.intersphinx_named_inventory = {}
+        inventories.clear()
+
         # Duplicate values in different inventories will shadow each
         # other; which one will override which can vary between builds
         # since they are specified using an unordered dict.  To make
@@ -289,46 +276,57 @@ def load_mappings(app):
         # add the unnamed inventories last.  This means that the
         # unnamed inventories will shadow the named ones but the named
         # ones can still be accessed when the name is specified.
-        cached_vals = list(cache.values())
+        cached_vals = list(inventories.cache.values())
         named_vals = sorted(v for v in cached_vals if v[0])
         unnamed_vals = [v for v in cached_vals if not v[0]]
         for name, _x, invdata in named_vals + unnamed_vals:
             if name:
-                env.intersphinx_named_inventory[name] = invdata
+                inventories.named_inventory[name] = invdata
             for type, objects in iteritems(invdata):
-                env.intersphinx_inventory.setdefault(
-                    type, {}).update(objects)
+                inventories.main_inventory.setdefault(type, {}).update(objects)
 
 
 def missing_reference(app, env, node, contnode):
+    # type: (Sphinx, BuildEnvironment, nodes.Node, nodes.Node) -> None
     """Attempt to resolve a missing reference via intersphinx references."""
     target = node['reftarget']
+    inventories = InventoryAdapter(env)
+    objtypes = None  # type: List[unicode]
     if node['reftype'] == 'any':
         # we search anything!
         objtypes = ['%s:%s' % (domain.name, objtype)
                     for domain in env.domains.values()
                     for objtype in domain.object_types]
         domain = None
-    elif node['reftype'] == 'doc':
-        domain = 'std'  # special case
-        objtypes = ['std:doc']
     else:
         domain = node.get('refdomain')
         if not domain:
             # only objects in domains are in the inventory
             return
-        objtypes = env.domains[domain].objtypes_for_role(node['reftype'])
+        objtypes = env.get_domain(domain).objtypes_for_role(node['reftype'])
         if not objtypes:
             return
         objtypes = ['%s:%s' % (domain, objtype) for objtype in objtypes]
-    to_try = [(env.intersphinx_inventory, target)]
+    if 'std:cmdoption' in objtypes:
+        # until Sphinx-1.6, cmdoptions are stored as std:option
+        objtypes.append('std:option')
+    to_try = [(inventories.main_inventory, target)]
+    if domain:
+        full_qualified_name = env.get_domain(domain).get_full_qualified_name(node)
+        if full_qualified_name:
+            to_try.append((inventories.main_inventory, full_qualified_name))
     in_set = None
     if ':' in target:
         # first part may be the foreign doc set name
         setname, newtarget = target.split(':', 1)
-        if setname in env.intersphinx_named_inventory:
+        if setname in inventories.named_inventory:
             in_set = setname
-            to_try.append((env.intersphinx_named_inventory[setname], newtarget))
+            to_try.append((inventories.named_inventory[setname], newtarget))
+            if domain:
+                node['reftarget'] = newtarget
+                full_qualified_name = env.get_domain(domain).get_full_qualified_name(node)
+                if full_qualified_name:
+                    to_try.append((inventories.named_inventory[setname], full_qualified_name))
     for inventory, target in to_try:
         for objtype in objtypes:
             if objtype not in inventory or target not in inventory[objtype]:
@@ -337,8 +335,11 @@ def missing_reference(app, env, node, contnode):
             if '://' not in uri and node.get('refdoc'):
                 # get correct path in case of subdirectories
                 uri = path.join(relative_path(node['refdoc'], '.'), uri)
-            newnode = nodes.reference('', '', internal=False, refuri=uri,
-                                      reftitle=_('(in %s v%s)') % (proj, version))
+            if version:
+                reftitle = _('(in %s v%s)') % (proj, version)
+            else:
+                reftitle = _('(in %s)') % (proj,)
+            newnode = nodes.reference('', '', internal=False, refuri=uri, reftitle=reftitle)
             if node.get('refexplicit'):
                 # use whatever title was given
                 newnode.append(contnode)
@@ -362,6 +363,7 @@ def missing_reference(app, env, node, contnode):
 
 
 def setup(app):
+    # type: (Sphinx) -> Dict[unicode, Any]
     app.add_config_value('intersphinx_mapping', {}, True)
     app.add_config_value('intersphinx_cache_limit', 5, False)
     app.add_config_value('intersphinx_timeout', None, False)
@@ -370,21 +372,38 @@ def setup(app):
     return {'version': sphinx.__display_version__, 'parallel_read_safe': True}
 
 
-if __name__ == '__main__':
-    # debug functionality to print out an inventory
-    import sys
+def debug(argv):
+    # type: (List[unicode]) -> None
+    """Debug functionality to print out an inventory"""
+    if len(argv) < 2:
+        print("Print out an inventory file.\n"
+              "Error: must specify local path or URL to an inventory file.",
+              file=sys.stderr)
+        sys.exit(1)
+
+    class MockConfig(object):
+        intersphinx_timeout = None  # type: int
+        tls_verify = False
 
     class MockApp(object):
         srcdir = ''
+        config = MockConfig()
 
         def warn(self, msg):
             print(msg, file=sys.stderr)
 
-    filename = sys.argv[1]
-    invdata = fetch_inventory(MockApp(), '', filename)
+    filename = argv[1]
+    invdata = fetch_inventory(MockApp(), '', filename)  # type: ignore
     for key in sorted(invdata or {}):
         print(key)
         for entry, einfo in sorted(invdata[key].items()):
             print('\t%-40s %s%s' % (entry,
                                     einfo[3] != '-' and '%-40s: ' % einfo[3] or '',
                                     einfo[2]))
+
+
+if __name__ == '__main__':
+    import logging  # type: ignore
+    logging.basicConfig()
+
+    debug(argv=sys.argv[1:])  # type: ignore
